#######################################################################
#                  cons2prim a la Kastaun et al 2021                  #
#######################################################################


@with_signature [simd=false] function cons2prim_valencia1d_freeze_flags(eq::AbstractEquation)
  @accepts D, Sr, τ, r
  @accepts c2p_reset_atm, c2p_limit_vr, c2p_freeze_atm
  @accepts γrr, sqrt_detγ
  D, Sr, _, τ, ρ, vr, _, v, ϵ, p, ρhW2, c2p_reset_ϵ, c2p_reset_atm, c2p_limit_vr, c2p_init_admissible, _ =
    cons2prim_impl(eq.eos, eq.cold_eos, eq, D, Sr, 0.0, τ, r, 0.0, γrr, 1.0, c2p_reset_atm, c2p_limit_vr, c2p_freeze_atm)
  rD  = sqrt_detγ*D
  rSr = sqrt_detγ*Sr
  rτ  = sqrt_detγ*τ
  @returns D, Sr, τ, rD, rSr, rτ, ρ, vr, v, ϵ, p, ρhW2, c2p_reset_ϵ, c2p_reset_atm, c2p_limit_vr, c2p_init_admissible
end
@with_signature [simd=false] function cons2prim_valencia1d(eq::AbstractEquation)
  @accepts D, Sr, τ, r
  @accepts c2p_reset_atm, c2p_limit_vr, c2p_freeze_atm
  @accepts γrr, sqrt_detγ
  D, Sr, _, τ, ρ, vr, _, v, ϵ, p, ρhW2, c2p_reset_ϵ, c2p_reset_atm, c2p_limit_vr, c2p_init_admissible, c2p_freeze_atm =
    cons2prim_impl(eq.eos, eq.cold_eos, eq, D, Sr, 0.0, τ, r, 0.0, γrr, 1.0, c2p_reset_atm, c2p_limit_vr, c2p_freeze_atm)
  rD  = sqrt_detγ*D
  rSr = sqrt_detγ*Sr
  rτ  = sqrt_detγ*τ
  @returns D, Sr, τ, rD, rSr, rτ, ρ, vr, v, ϵ, p, ρhW2, c2p_reset_ϵ, c2p_reset_atm, c2p_limit_vr, c2p_init_admissible, c2p_freeze_atm
end


@with_signature [simd=false] function cons2prim_spherical1d_freeze_flags(eq::AbstractEquation)
  @accepts D, Sr, τ, r
  @accepts c2p_reset_atm, c2p_limit_vr, c2p_freeze_atm
  @accepts γrr
  D, Sr, _, τ, ρ, vr, _, v, ϵ, p, ρhW2, c2p_reset_ϵ, c2p_reset_atm, c2p_limit_vr, c2p_init_admissible, _ =
    cons2prim_impl(eq.eos, eq.cold_eos, eq, D, Sr, 0.0, τ, r, 0.0, γrr, 1.0, c2p_reset_atm, c2p_limit_vr, c2p_freeze_atm)
  @returns D, Sr, τ, ρ, vr, v, ϵ, p, ρhW2, c2p_reset_ϵ, c2p_reset_atm, c2p_limit_vr, c2p_init_admissible
end

@with_signature [simd=false] function cons2prim_spherical1d(eq::AbstractEquation)
  @accepts D, Sr, τ, r
  @accepts c2p_reset_atm, c2p_limit_vr, c2p_freeze_atm
  @accepts γrr
  D, Sr, _, τ, ρ, vr, _, v, ϵ, p, ρhW2, c2p_reset_ϵ, c2p_reset_atm, c2p_limit_vr, c2p_init_admissible, c2p_freeze_atm =
    cons2prim_impl(eq.eos, eq.cold_eos, eq, D, Sr, 0.0, τ, r, 0.0, γrr, 1.0, c2p_reset_atm, c2p_limit_vr, c2p_freeze_atm)
  @returns D, Sr, τ, ρ, vr, v, ϵ, p, ρhW2, c2p_reset_ϵ, c2p_reset_atm, c2p_limit_vr, c2p_init_admissible, c2p_freeze_atm
end


@with_signature [simd=false] function cons2prim_rescaled_spherical1d_freeze_flags(eq::AbstractEquation)
  @accepts D, Sr, τ, r
  @accepts c2p_reset_atm, c2p_limit_vr, c2p_freeze_atm
  @accepts γrr, B
  D, Sr, _, τ, ρ, vr, _, v, ϵ, p, ρhW2, c2p_reset_ϵ, c2p_reset_atm, c2p_limit_vr, c2p_init_admissible, _ =
    cons2prim_impl(eq.eos, eq.cold_eos, eq, D, Sr, 0.0, τ, r, 0.0, γrr, 1.0, c2p_reset_atm, c2p_limit_vr, c2p_freeze_atm)
  # in case something was adjusted by cons2prim
  rD  = r*B^3*D
  rSr = r*B^3*Sr
  rτ  = r*B^3*τ
  @returns D, Sr, τ, rD, rSr, rτ, ρ, vr, v, ϵ, p, ρhW2, c2p_reset_ϵ, c2p_reset_atm, c2p_limit_vr, c2p_init_admissible
end
@with_signature [simd=false] function cons2prim_rescaled_spherical1d(eq::AbstractEquation)
  @accepts D, Sr, τ, r
  @accepts c2p_reset_atm, c2p_limit_vr, c2p_freeze_atm
  @accepts γrr, B
  D, Sr, _, τ, ρ, vr, _, v, ϵ, p, ρhW2, c2p_reset_ϵ, c2p_reset_atm, c2p_limit_vr, c2p_init_admissible, c2p_freeze_atm =
    cons2prim_impl(eq.eos, eq.cold_eos, eq, D, Sr, 0.0, τ, r, 0.0, γrr, 1.0, c2p_reset_atm, c2p_limit_vr, c2p_freeze_atm)
  # in case something was adjusted by cons2prim
  rD  = r*B^3*D
  rSr = r*B^3*Sr
  rτ  = r*B^3*τ
  @returns D, Sr, τ, rD, rSr, rτ, ρ, vr, v, ϵ, p, ρhW2, c2p_reset_ϵ, c2p_reset_atm, c2p_limit_vr, c2p_init_admissible, c2p_freeze_atm
end


@with_signature [simd=false] function cons2prim_doublecartoon_freeze_flags(eq::AbstractEquation)
  @accepts D, Sx, τ, r
  @accepts c2p_reset_atm, c2p_limit_vr, c2p_freeze_atm
  @accepts γxx, sqrtdetγ
  D, Sx, _, τ, ρ, vx, _, v, ϵ, p, ρhW2, c2p_reset_ϵ, c2p_reset_atm, c2p_limit_vr, c2p_init_admissible, _ =
    cons2prim_impl(eq.eos, eq.cold_eos, eq, D, Sx, 0.0, τ, r, 0.0, γxx, 1.0, c2p_reset_atm, c2p_limit_vr, c2p_freeze_atm)
  # in case something was adjusted by cons2prim
  rD  = sqrtdetγ*D
  rSx = sqrtdetγ*Sx
  rτ  = sqrtdetγ*τ
  @returns D, Sx, τ, rD, rSx, rτ, ρ, vx, v, ϵ, p, ρhW2, c2p_reset_ϵ, c2p_reset_atm, c2p_limit_vr, c2p_init_admissible
end
@with_signature [simd=false] function cons2prim_doublecartoon(eq::AbstractEquation)
  @accepts D, Sx, τ, r
  @accepts c2p_reset_atm, c2p_limit_vr, c2p_freeze_atm
  @accepts γxx, sqrtdetγ
  D, Sx, _, τ, ρ, vx, _, v, ϵ, p, ρhW2, c2p_reset_ϵ, c2p_reset_atm, c2p_limit_vr, c2p_init_admissible, c2p_freeze_atm =
    cons2prim_impl(eq.eos, eq.cold_eos, eq, D, Sx, 0.0, τ, r, 0.0, γxx, 1.0, c2p_reset_atm, c2p_limit_vr, c2p_freeze_atm)
  # in case something was adjusted by cons2prim
  rD  = sqrtdetγ*D
  rSx = sqrtdetγ*Sx
  rτ  = sqrtdetγ*τ
  @returns D, Sx, τ, rD, rSx, rτ, ρ, vx, v, ϵ, p, ρhW2, c2p_reset_ϵ, c2p_reset_atm, c2p_limit_vr, c2p_init_admissible, c2p_freeze_atm
end


@with_signature [simd=false] function cons2prim_cartoon_freeze_flags(eq::AbstractEquation)
  @accepts D, Sx, Sz, τ, x, y
  @accepts c2p_reset_atm, c2p_limit_vr, c2p_freeze_atm
  @accepts γxx, γzz, sqrtdetγ
  z = y
  D, Sx, Sz, τ, ρ, vx, vz, v, ϵ, p, ρhW2, c2p_reset_ϵ, c2p_reset_atm, c2p_limit_vr, c2p_init_admissible, _ =
    cons2prim_impl(eq.eos, eq.cold_eos, eq, D, Sx, Sz, τ, x, z, γxx, γzz, c2p_reset_atm, c2p_limit_vr, c2p_freeze_atm)
  # in case something was adjusted by cons2prim
  rD  = sqrtdetγ*D
  rSx = sqrtdetγ*Sx
  rSz = sqrtdetγ*Sz
  rτ  = sqrtdetγ*τ
  @returns D, Sx, Sz, τ, rD, rSx, rSz, rτ, ρ, vx, vz, v, ϵ, p, ρhW2, c2p_reset_ϵ, c2p_reset_atm, c2p_limit_vr, c2p_init_admissible
end
@with_signature [simd=false] function cons2prim_cartoon(eq::AbstractEquation)
  @accepts D, Sx, Sz, τ, x, y
  @accepts c2p_reset_atm, c2p_limit_vr, c2p_freeze_atm
  @accepts γxx, γzz, sqrtdetγ
  z = y
  D, Sx, Sz, τ, ρ, vx, vz, v, ϵ, p, ρhW2, c2p_reset_ϵ, c2p_reset_atm, c2p_limit_vr, c2p_init_admissible, c2p_freeze_atm =
    cons2prim_impl(eq.eos, eq.cold_eos, eq, D, Sx, Sz, τ, x, z, γxx, γzz, c2p_reset_atm, c2p_limit_vr, c2p_freeze_atm)
  # in case something was adjusted by cons2prim
  rD  = sqrtdetγ*D
  rSx = sqrtdetγ*Sx
  rSz = sqrtdetγ*Sz
  rτ  = sqrtdetγ*τ
  @returns D, Sx, Sz, τ, rD, rSx, rSz, rτ, ρ, vx, vz, v, ϵ, p, ρhW2, c2p_reset_ϵ, c2p_reset_atm, c2p_limit_vr, c2p_init_admissible, c2p_freeze_atm
end


# reconstruction is covariant, so for spherical symmetry just set the z components to zero
# cons2prim from Kastaun et al. arXiv:2005.01821
@inline function cons2prim_impl(
    eos::AbstractEquationOfState, _cold_eos::ColdEquationOfState,
    parameters,
    D, Sx, Sz, τ,
    xcoord, zcoord, γxx, γzz,
    c2p_reset_atm, c2p_limit_vr, c2p_freeze_atm)

  verbose = false
  c2p_reset_ϵ, c2p_init_admissible = 0.0, 1.0
  # setup some constants
  @unpack atm_density, atm_threshold,
          c2p_threshold_velocity_limiter, c2p_set_atmosphere_on_failure = parameters
  # ignore _cold_eos when eos is already a cold one
  cold_eos = eos isa ColdEquationOfState ? eos : _cold_eos
  # bounds on EoS, assuming cold EoS
  ρmin = atm_density * atm_threshold
  pmin = cold_eos(Pressure, Density, ρmin)
  ϵmin = cold_eos(InternalEnergy, Density, ρmin)
  Dmin = ρmin * 1.0 # W = 1
  ρmax = Inf
  # atmosphere values, assuming cold EoS
  ρatm = atm_density
  patm = cold_eos(Pressure, Density, ρatm)
  ϵatm = cold_eos(InternalEnergy, Density, ρatm)
  # lower bound on enthalpy, although h ≥ 1 is physical bound
  h0 = 1e-3
  # parameters for Dumbser+ 2024 velocity filter
  y0 = c2p_threshold_velocity_limiter
  e0 = y0^2/2.0 # fudge factor for filtered division in velocity limiter
  # assuming diagonal metric
  γuuxx = 1.0/γxx
  γuuzz = 1.0/γzz

  if (c2p_freeze_atm > 0.0 && c2p_reset_atm == 0.0 && D < ρmin)
    @warn "Require atmosphere but denied by mask" D ρmin xcoord zcoord c2p_freeze_atm c2p_reset_atm
    TODO("Upsi")
  elseif (c2p_freeze_atm == 0.0 && D < ρmin) || (c2p_freeze_atm > 0.0 && c2p_reset_atm > 0.0)
    if D < ρmin
      c2p_init_admissible = 0.0
    end
    @label impose_atmosphere
    ρ  = ρatm
    vx = 0.0
    vz = 0.0
    v  = 0.0
    W  = 1.0
    p  = patm
    ϵ  = ϵatm
    c2p_reset_atm = 1.0
  else
    # (22)-(23), Si := S_i is 3-covector
    q = τ / D
    rx = Sx / D
    rz = Sz / D
    r2 = γuuxx*rx*rx+γuuzz*rz*rz
    r  = sqrt(r2)
    k = r / (1.0+q)
    if !(0.0 ≤ k ≤ 1.0) # admissibility constraint from arXiv:1306.4953, Appendix
      c2p_init_admissible = 1.0
      # rescale Si such that r = 1+q
      rescale = 0.99*(1+q)/r
      Sx *= rescale
      Sz *= rescale
      rx = Sx / D
      rz = Sz / D
      r2 = γuuxx*rx*rx+γuuzz*rz*rz
      r  = sqrt(r2)
    end
    z0 = r/h0
    v0 = z0 / sqrt(1.0 + z0^2)
    h02 = h0^2
    # setup anonymous functions for solvers
    limit_v = let D=D, v0=v0, Sx=Sx, Sz=Sz, r=r, γuuxx=γuuxx, γuuzz=γuuzz, _y0=y0, _e0=e0
      # infamous closure bug ...
      # https://docs.julialang.org/en/v1/manual/performance-tips/#man-performance-captured
      mu -> begin
        _y = D/mu # =^= ρhW^2
        # TODO Make this method return the components, so that we can also invoke it later on
        _v = if _y > _y0
          min(mu*r, v0)
        else
          # limit velocity according to Dumbser et al arxiv:2307.06629
          _fy = 2.0 * (_y/_y0)^3 - 3.0 * (_y/_y0)^2 + 1.0
          _f = _y / (_y^2 + _fy * _e0)
          # no need to use rx*D here, because D,rx,Sx should remain unchanged throughout
          _vx = Sx * _f
          _vz = Sz * _f
          sqrt(γuuxx*_vx*_vx+γuuzz*_vz*_vz)
        end
        return _v
      end
    end
    # What is not a question but rather W^hat, cf. (40)
    fn_What = let r=r, v0=v0, limit_v=limit_v
      mu -> begin
        vhat = limit_v(mu)
        return 1.0 / sqrt(1.0 - vhat^2)
      end
    end
    # infamous closure bug ...
    # https://docs.julialang.org/en/v1/manual/performance-tips/#man-performance-captured
    master_fn = let D=D, τ=τ, q=q, r2=r2, limit_v=limit_v, ρmin=ρmin, ϵmin=ϵmin, eos=eos, _y0=y0, cold_eos=cold_eos,
                    ρatm=ρatm, patm=patm, ϵatm=ϵatm
      # Note: In the GRMHD case r and q would depend on μ! See (26), (30), (31)
      # Here we only consider GRHD where we put Bᵢ = bᵢ = 0.
      # (25) says that when Bᵢ = 0 the decomposition of rᵢ into parallel and normal
      # parts is undefined. Nevertheless, one can show that in this case there is just
      # no parallel part, I think.
      mu -> begin # eq (44)
        _vhat = limit_v(mu)
        _What = 1.0 / sqrt(1.0 - _vhat^2)
        ρhat = D / _What
        if _vhat > v0
          _vhat = v0
          _What = 1.0 / sqrt(1.0 - _vhat^2)
          ρhat = D / _What
        end
        if ρhat < ρmin
          ρhat, phat, ϵhat = ρatm, patm, ϵatm
        else
          _y = D/mu
          # compute thermodynamic vars
          if eos isa ColdEquationOfState || _y ≤ _y0
            phat = cold_eos(Pressure, Density, ρhat)
            ϵhat = cold_eos(InternalEnergy, Density, ρhat)
          else
            # eq (42)
            ϵhat = _What * (q - mu * r2) + _vhat^2 * _What^2 / (1.0 + _What)
            ϵhat = max(ϵhat, ϵatm) # enforce EOS bounds
            phat = eos(Pressure, Density, ρhat, InternalEnergy, ϵhat)
          end
        end
        ahat = phat / (ρhat * (1.0 + ϵhat))
        # eq (46)-(48)
        nu_A = (1.0 + ahat) * (1.0 + ϵhat) / _What
        nu_B = (1.0 + ahat) * (1.0 + q - mu * r2)
        nu = max(nu_A, nu_B)
        # eq (44)
        return mu - 1.0 / (nu + r2 * mu)
      end
    end

    # setup bracket
    mu_a = 0.0
    # with no EM fields we can find the root of (49) analytically, because r = const.
    mu_plus = 1.0 / sqrt(h02 + r2)
    mu_b = r < h0 ? 1.0 / h0 : mu_plus
    What_a, What_b, What_plus = fn_What(mu_a), fn_What(mu_b), fn_What(mu_plus)
    if D / What_plus > ρmax
      error("Invalid state encountered!")
    end

    # upper root bracket adjustment, cf. Appendix
    if D / What_plus < ρmax < D
      verbose && @info "adjusting upper bracket"
      # infamous closure bug ...
      # https://docs.julialang.org/en/v1/manual/performance-tips/#man-performance-captured
      fn_ρmax = let D=D, ρmax=ρmax, fn_What=fn_What
        mu -> D / fn_What(mu) - ρmax
      end
      mu_b = find_zero(fn_ρmax, (mu_a, mu_plus), Roots.Brent())
      # mu_b = Roots.a42(fn_ρmax, (mu_a, mu_plus))
      if isnan(mu_b)
        error("Failed to correct upper interval bound")
      end
      What_b = fn_What(mu_b)
    end
    # lower root bracket adjustment, cf. Appendix
    if D / What_plus < ρmin < D # yes, we need What_b here; TODO Why don't we use What_b now?
      verbose && @info "adjusting lower bracket"
      # infamous closure bug ...
      # https://docs.julialang.org/en/v1/manual/performance-tips/#man-performance-captured
      fn_ρmin = let D=D, ρmin=ρmin, fn_What=fn_What
        mu -> D / fn_What(mu) - ρmin
      end
      mu_a = find_zero(fn_ρmin, (mu_a, mu_b), Roots.Brent())
      # mu_a = Roots.a42(fn_ρmin, (mu_a, mu_plus))
      What_a = fn_What(mu_a)
    end
    if verbose
      # verify if there is a solution inside bracket
      mf_a, mf_b = master_fn(mu_a), master_fn(mu_b)
      mf_ab = mf_a*mf_b
      k = r/(1.0+q)
      if (mf_ab > 0.0 || !isfinite(mf_ab))
        @warn "invalid bracket" xcoord zcoord (D,Sx,Sz,τ) (mu_a, mu_b, mu_plus) (mf_a, mf_b, mf_ab) (What_a, What_b, mu_a*r, mu_b*r, D/mu_a, D/mu_b, q, k)
      end
    end

    # now solve the equation
    if c2p_set_atmosphere_on_failure
      mu = solve(ZeroProblem(master_fn,(mu_a, mu_b)), Roots.Brent())
      !isfinite(mu) && @goto impose_atmosphere
    else
      mu = find_zero(master_fn, (mu_a, mu_b), Roots.Brent())
    end
    # mu = Roots.a42(master_fn, (mu_a, mu_b))

    # TODO Rename c2p_limit_vr to c2p_limit_v
    # now limit the velocity components instead of the absolute value,
    # limiting according to Dumbser et al arxiv:2307.06629
    y = D/mu # =^= ρhW^2
    if y > y0
      vx = min(mu*rx, v0)
      vz = min(mu*rz, v0)
    else
      c2p_limit_vr = 1.0
      fy = 2.0 * (y/y0)^3 - 3.0 * (y/y0)^2 + 1.0
      f = y / (y^2 + fy * e0)
      vx = Sx * f
      vz = Sz * f
    end
    v2 = γuuxx*vx*vx+γuuzz*vz*vz
    v = sqrt(v2)
    W = 1.0 / sqrt(1.0 - v2)
    ρ = D / W
    if v > v0
      vx = vx*v0/v
      vz = vz*v0/v
      v = v0
      W = 1.0 / sqrt(1.0 - v^2)
      ρ = D / W
    end
    if (c2p_freeze_atm > 0.0 && c2p_reset_atm == 0.0 && ρ < ρmin)
      @warn "Require atmosphere but denied by mask" D ρmin xcoord zcoord c2p_freeze_atm c2p_reset_atm
      TODO("Upsi")
    elseif (c2p_freeze_atm == 0.0 && ρ < ρmin) || (c2p_freeze_atm > 0.0 && c2p_reset_atm > 0.0)
      @goto impose_atmosphere
    end
    # compute thermodynamic vars
    if eos isa ColdEquationOfState || y ≤ y0
      ϵ = cold_eos(InternalEnergy, Density, ρ)
      p = cold_eos(Pressure, Density, ρ)
    else
      # eq (42)
      ϵ = W * (q - mu * r2) + v2 * W^2 / (1.0 + W)
      ϵ = max(ϵ, ϵatm) # enforce EOS bounds
      p = eos(Pressure, Density, ρ, InternalEnergy, ϵ)
    end
  end

  h    = 1.0+ϵ+p/ρ
  W2   = 1.0/(1.0-v^2)
  ρhW2 = ρ*h*W2
  # recomputing conservatives in case we limited something
  D  = W*ρ
  Sx = ρhW2*vx
  Sz = ρhW2*vz
  τ  = ρhW2 - p - D

  return D, Sx, Sz, τ, ρ, vx, vz, v, ϵ, p, ρhW2,
         c2p_reset_ϵ, c2p_reset_atm, c2p_limit_vr, c2p_init_admissible, c2p_freeze_atm
end


@with_signature function determine_atmosphere(equation::Equation)
  @accepts D
  @unpack atm_density, atm_threshold, eos = equation
  ρmin = atm_density * atm_threshold
  c2p_reset_atm = Float64(D<ρmin)
  @returns c2p_reset_atm
end
